/* Copyright 2005 Jem Berkes
 * www.sysdesign.ca
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "ap_config.h"
#include "apr_optional.h"
#include "apr_hash.h"
#include "apr_strings.h"
#include "httpd.h"
#include "http_config.h"
#include "http_connection.h"
#include "util_filter.h"

#include "dnsbl_lookup.h"


/* Link between httpd and dnsbl_lookup module */
module AP_MODULE_DECLARE_DATA dnsbl_lookup_module;


/* Module configuration, on a per-server basis (virtual host container) */
typedef struct dnsbl_lookup_cfg
{
    int lookup_enabled;
    apr_hash_t* ipv4_dnsbl_chains;	/* many chains can be held inside these */
/*	apr_hash_t* ipv6_dnsbl_chains;	not supported currently, but easily added */
    apr_hash_t* rhsbl_chains;
} dnsbl_lookup_cfg;



char* dns_lookup_host(const char* hostname, apr_pool_t* p)
{
    apr_sockaddr_t* sa;
    char* dnsresult;
    if ((apr_sockaddr_info_get(&sa, hostname, APR_INET, 0, 0, p) == APR_SUCCESS) &&
        (apr_sockaddr_ip_get(&dnsresult, sa) == APR_SUCCESS))
        return dnsresult;
    else
        return NULL;
}


/*
    Converts IPv4 address to reverse octet notation (NULL on error)
    TODO: more elegant binary based reversal? that is reliably portable?
*/
char* ipreverse(apr_pool_t* p, apr_sockaddr_t* address)
{
    char* ipstring;
    if (apr_sockaddr_ip_get(&ipstring, address) == APR_SUCCESS) {
        char *token, *tokstate;
        int n;
        int octets[4] = {0, 0, 0, 0};
        token = apr_strtok(ipstring, ".", &tokstate);
        for (n=0; token && (n<4); n++) {
            octets[n] = atoi(token);
            token = apr_strtok(NULL, ".", &tokstate);
        }
        return apr_psprintf(p, "%d.%d.%d.%d", octets[3], octets[2], octets[1], octets[0]);
    }
    return NULL;
}


/*
    This function does the work for dnsbl_lookup_ip() and _domain(). They
    provide us the appropriate chain (IPv4 DNSBL, IPv6 DNSBL, RHSBL) and
    a text form of the key to lookup. Then the queries always work the same:
    
    - Go down the servers under the indicated chain of zones
    - Do a DNS lookup on key.zonedomain
    - Analyze response and act depending on querymode
    
    Note that zonedata may be NULL if user doesn't want storage of results.
    Returns DNSBL_POSITIVE, DNSBL_NEGATIVE, or DNSBL_FAILURE
*/
static int dnsbl_lookup(apr_array_header_t* zones, int querymode, const char* key,
    apr_pool_t* p, apr_table_t* zonedata, dnsbl_lookup_cfg* cfg)
{
    int zonenum, numpositive = 0;
    int numservers = zones->nelts;
    zone_entry* servers = (zone_entry *)zones->elts;

    if ((querymode!=DNSBL_ANYPOSTV_RETFIRST) && (querymode!=DNSBL_ANYPOSTV_RETEVERY)
         && (querymode!=DNSBL_ALLPOSTV_RETEVERY))
         return DNSBL_FAILURE;	/* invalid querymode */

    for (zonenum=0; zonenum < numservers; zonenum++) {
        const char* resultaddr =
            dns_lookup_host(apr_pstrcat(p, key, ".", servers[zonenum].zonedomain, NULL), p);
        if (resultaddr) {
            /* we got some result from the DNS query, not NXDOMAIN. should we consider 'positive'? */
            const char* postvresp = servers[zonenum].postvresp;	/* check positive match criteria */
            if ((apr_strnatcasecmp(postvresp, DNSBL_MATCH_ANY) == 0) ||
                (strncmp(resultaddr, postvresp, strlen(postvresp)) == 0)) {
                numpositive++;
                if (zonedata)
                    apr_table_add(zonedata, servers[zonenum].zonedomain, resultaddr);
                if (querymode == DNSBL_ANYPOSTV_RETFIRST)
                    return DNSBL_POSITIVE;	/* found one positive, returning single */
            }
        }
    }
    /* all servers were queried */
    if (numpositive == numservers)
        return DNSBL_POSITIVE;
    else if ((querymode == DNSBL_ANYPOSTV_RETEVERY) && (numpositive > 0))
        return DNSBL_POSITIVE;
    else
        return DNSBL_NEGATIVE;
}


/*

Do DNSBL lookups for an IP address

chain	The chain name from the configuration to query

querymode
DNSBL_ANYPOSTV_RETFIRST - Any positive from chain, stop and return first
DNSBL_ANYPOSTV_RETEVERY - Any positive, check all and return every positive
DNSBL_ALLPOSTV_RETEVERY - All must check positive, return every positive

address	The address to look up. Only IPv4 is currently supported.

p	Pool to use for storage

s	Server record whose module configuration to use

zonedata (output)
If zonedata is NULL then there will be none returned. Otherwise, a new table
will be created and filled with {zone, result} pairs for each zone that
comes back positive. The result is an IPv4 string, probably 127.0.0.x. Note
that this zonedata table will only contain entries if there is some positive
match, i.e., only if the return code was DNSBL_POSITIVE.

Returns:
DNSBL_POSITIVE - Positive match (zonedata has details, if requested)
DNSBL_NEGATIVE - Negative
DNSBL_FAILURE - Generic failure, e.g. DnsblLookups Off or invalid chain
	
*/
static int dnsbl_lookup_ip(const char* chain, int querymode, apr_sockaddr_t* address,
    apr_pool_t* p, server_rec* s, apr_table_t** zonedata)
{
    dnsbl_lookup_cfg* cfg;
    apr_table_t* zones = NULL;	/* might not return zonedata */
    apr_array_header_t* zonechain;
    char* lookupkey;

    cfg = ap_get_module_config(s->module_config, &dnsbl_lookup_module);
    if (!(cfg->lookup_enabled))
        return DNSBL_FAILURE;	/* DNSBL lookups are not enabled */
    if (!address)
        return DNSBL_FAILURE;	/* null address given */
    /*
        Require IPv4. TODO: add IPv6 support
        That would involve apr_hash_get from ipv6_dnsbl_chains and
        preparing an appropriate lookupkey, string form of IPv6
    */
    if (address->ipaddr_len > 16)
        return DNSBL_FAILURE;
    lookupkey = ipreverse(p, address);
    if (!lookupkey)
        return DNSBL_FAILURE;	/* unable to prepare lookup string from address */
    zonechain = (apr_array_header_t *)apr_hash_get(cfg->ipv4_dnsbl_chains, chain, APR_HASH_KEY_STRING);
    if (!zonechain)
        return DNSBL_FAILURE;	/* no such chain */
    if (zonedata) {
        *zonedata = apr_table_make(p, 1);
        zones = *zonedata;
    }
    /* pass work to main lookup function */
    return dnsbl_lookup(zonechain, querymode, lookupkey, p, zones, cfg);
}


/*
    Similar to dnsbl_lookup_ip() except _domain is used for RHSBLs to query some
    host or domain name. Return codes the same - passes through to dnsbl_lookup()
*/
static int dnsbl_lookup_domain(const char* chain, int querymode, const char* domain,
    apr_pool_t* p, server_rec* s, apr_table_t** zonedata)
{
    dnsbl_lookup_cfg* cfg;
    apr_table_t* zones = NULL;	/* might not return zonedata */
    apr_array_header_t* zonechain;

    cfg = ap_get_module_config(s->module_config, &dnsbl_lookup_module);
    if (!(cfg->lookup_enabled))
        return DNSBL_FAILURE;	/* DNSBL lookups are not enabled */
    if (!domain)
        return DNSBL_FAILURE;	/* null domain given */

    zonechain = (apr_array_header_t *)apr_hash_get(cfg->rhsbl_chains, chain, APR_HASH_KEY_STRING);
    if (!zonechain)
        return DNSBL_FAILURE;	/* no such chain */
    if (zonedata) {
        *zonedata = apr_table_make(p, 1);
        zones = *zonedata;
    }
    /* pass work to main lookup function */
    return dnsbl_lookup(zonechain, querymode, domain, p, zones, cfg);
}


/*
    Called by Apache to create per-server configuration (virtual host)
    Returns pointer to module-specific structure
*/
static void* dnsbl_lookup_create_server_config(apr_pool_t* p, server_rec* s)
{
    dnsbl_lookup_cfg* cfg = apr_pcalloc(p, sizeof(dnsbl_lookup_cfg));
    cfg->lookup_enabled = 0;	/* disabled by default */
    cfg->ipv4_dnsbl_chains = apr_hash_make(p);
    cfg->rhsbl_chains = apr_hash_make(p);
    return cfg;
}


static const char* enable_dnsbl_lookups(cmd_parms* cmd, void* dummy, int arg)
{
    dnsbl_lookup_cfg* cfg = ap_get_module_config(cmd->server->module_config, &dnsbl_lookup_module);
    if (cfg)
        cfg->lookup_enabled = arg;
    return NULL;
}

/*
    Helper function to add DNSBL and RHSBL zones into chains.
    Hash tables store the chain names under each chain namespace
    The hash table entries are arrays of zone_entry structs
*/
static const char* add_into_chains(apr_pool_t* p, apr_hash_t* chainspace,
    const char* chain, const char* domain, const char* postvresp)
{
    apr_array_header_t* zonechain;
    zone_entry* newentry;

    zonechain = (apr_array_header_t *)apr_hash_get(chainspace, chain, APR_HASH_KEY_STRING);
    if (!zonechain) {	/* this chain does not exist yet */
        zonechain = apr_array_make(p, 1, sizeof(zone_entry));
        apr_hash_set(chainspace, chain, APR_HASH_KEY_STRING, zonechain);
    }
    newentry = (zone_entry *) apr_array_push(zonechain);
    newentry->zonedomain = domain;
    newentry->postvresp = postvresp;
    return NULL;
}


static const char* set_ipv4_dnsbl(cmd_parms* cmd, void* dummy, const char* chain,
    const char* domain, const char* postvresp)
{
    dnsbl_lookup_cfg* cfg = ap_get_module_config(cmd->server->module_config, &dnsbl_lookup_module);
    return add_into_chains(cmd->pool, cfg->ipv4_dnsbl_chains, chain, domain, postvresp);
}


static const char* set_rhsbl(cmd_parms* cmd, void* dummy, const char* chain,
    const char* domain, const char* postvresp)
{
    dnsbl_lookup_cfg* cfg = ap_get_module_config(cmd->server->module_config, &dnsbl_lookup_module);
    return add_into_chains(cmd->pool, cfg->rhsbl_chains, chain, domain, postvresp);	
}

static void dnsbl_lookup_register_hooks(apr_pool_t *p)
{
    APR_REGISTER_OPTIONAL_FN(dnsbl_lookup_ip);	/* Register our utility function (optional function) */
    APR_REGISTER_OPTIONAL_FN(dnsbl_lookup_domain);
}


static const command_rec dnsbl_lookup_cmds[] = {
    AP_INIT_FLAG("DnsblLookups", enable_dnsbl_lookups, NULL, RSRC_CONF, "Enable DNSBL and RHSBL lookups"),
    AP_INIT_TAKE3("DnsblZone", set_ipv4_dnsbl, NULL, RSRC_CONF, "Define an IPv4 DNSBL zone in a chain"),
    AP_INIT_TAKE3("RhsblZone", set_rhsbl, NULL, RSRC_CONF, "Define a RHSBL zone in a chain"),
    { NULL }
};


module AP_MODULE_DECLARE_DATA dnsbl_lookup_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    dnsbl_lookup_create_server_config,
    NULL,
    dnsbl_lookup_cmds,
    dnsbl_lookup_register_hooks
};
